/*
 * $Id$
 * $URL$
 * ----------------------------------------------------------------------------
 * This file is part of JSBML. Please visit <http://sbml.org/Software/JSBML>
 * for the latest version of JSBML and more information about SBML.
 *
 * Copyright (C) 2009-2011 jointly by the following organizations:
 * 1. The University of Tuebingen, Germany
 * 2. EMBL European Bioinformatics Institute (EBML-EBI), Hinxton, UK
 * 3. The California Institute of Technology, Pasadena, CA, USA
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation. A copy of the license agreement is provided
 * in the file named "LICENSE.txt" included with this software distribution
 * and also available online as <http://sbml.org/Software/JSBML/License>.
 * ----------------------------------------------------------------------------
 */

package org.sbml.jsbml;

import java.util.HashSet;
import java.util.Set;

import org.sbml.jsbml.CVTerm.Qualifier;
import org.sbml.jsbml.ListOf.Type;
import org.sbml.jsbml.Unit.Kind;
import org.sbml.jsbml.util.StringTools;

/**
 * Represents the unitDefinition XML element of a SBML file.
 * 
 * @author Andreas Dr&auml;ger
 * @author marine
 */
public class UnitDefinition extends AbstractNamedSBase {

	/**
	 * Generated serial version identifier.
	 */
	private static final long serialVersionUID = -4705380036260408123L;

	/**
	 * Predefined unit for area.
	 */
	public static final UnitDefinition area(int level, int version) {
		return getPredefinedUnit("area", level, version);
	}
	
	/**
	 * <p>
	 * Predicate returning <code>true</code> or <code>false</code> depending on
	 * whether two UnitDefinition objects are compatible.
	 * </p>
	 * <p>
	 * For the purposes of performing this comparison, two
	 * {@link UnitDefinition} objects are considered compatible when they
	 * contain compatible list of {@link Unit} objects. This means two
	 * {@link UnitDefinition} objects are compatible if both satisfy the method
	 * {@link #areEquivalent(UnitDefinition, UnitDefinition)} or one of both has
	 * {@link Kind#INVALID} as {@link Unit.Kind}
	 * 
	 * @param ud1
	 *            the first {@link UnitDefinition} object to compare
	 * @param ud2
	 *            the second {@link UnitDefinition} object to compare
	 * @return <code>true</code> if all the {@link Unit} objects in ud1 are
	 *         compatible to the {@link Unit} objects in ud2, <code>false</code>
	 *         otherwise.
	 * @see #areIdentical(UnitDefinition, UnitDefinition)
	 * @see Unit#areEquivalent(Unit, Unit)
	 */
	public static boolean areCompatible(UnitDefinition ud1, UnitDefinition ud2) {
		return areEquivalent(ud1, ud2) || ud2.isInvalid() || ud1.isInvalid();
	}

	/**
	 * Checks whether the given {@link UnitDefinition} and the
	 * {@link UnitDefinition} or {@link Unit} represented by the given
	 * {@link String} are equivalent.
	 * 
	 * @param ud
	 * @param units
	 * @return
	 * @see #areEquivalent(UnitDefinition, UnitDefinition)
	 * @see Unit#areEquivalent(Unit, Unit)
	 */
	public static boolean areEquivalent(UnitDefinition ud, String units) {
		UnitDefinition ud2 = ud.getModel().getUnitDefinition(units);
		if (ud2 != null) {
			return areEquivalent(ud, ud2);
		} else if (ud.isUnitKind()
				&& Unit.isUnitKind(units, ud.getLevel(), ud.getVersion())) {
			return Unit.areEquivalent(ud.getUnit(0), units);
		}
		
		return false;
	}

	/**
	 * <p>
	 * Predicate returning true or false depending on whether two UnitDefinition
	 * objects are equivalent.
	 * </p>
	 * <p>
	 * For the purposes of performing this comparison, two UnitDefinition
	 * objects are considered equivalent when they contain equivalent list of
	 * Unit objects. Unit objects are in turn considered equivalent if they
	 * satisfy the predicate Unit.areEquivalent(). The predicate tests a subset
	 * of the objects's attributes.
	 * </p>
	 * 
	 * @param ud1
	 *            the first UnitDefinition object to compare
	 * @param ud2
	 *            the second UnitDefinition object to compare
	 * @return true if all the Unit objects in ud1 are equivalent to the Unit
	 *         objects in ud2, false otherwise.
	 * @see areIdentical
	 * @see Unit.areEquivalent
	 */
	public static boolean areEquivalent(UnitDefinition ud1, UnitDefinition ud2) {
		UnitDefinition ud1clone = ud1.clone().simplify();
		UnitDefinition ud2clone = ud2.clone().simplify();
		if (ud1clone.getNumUnits() == ud2clone.getNumUnits()) {
			boolean equivalent = true;
			for (int i = 0; i < ud1clone.getNumUnits(); i++) {
				equivalent &= Unit.areEquivalent(ud1clone.getUnit(i), ud2clone
						.getUnit(i));
			}
			return equivalent;
		}
		return false;
	}

	/**
	 * <p>
	 * Predicate returning true or false depending on whether two UnitDefinition
	 * objects are identical.
	 * </p>
	 * <p>
	 * For the purposes of performing this comparison, two UnitDefinition
	 * objects are considered identical when they contain identical lists of
	 * Unit objects. Pairs of Unit objects in the lists are in turn considered
	 * identical if they satisfy the predicate Unit.areIdentical(). The
	 * predicate compares every attribute of the Unit objects.
	 * </p>
	 * 
	 * @param ud1
	 *            the first UnitDefinition object to compare
	 * @param ud2
	 *            the second UnitDefinition object to compare
	 * @return true if all the Unit objects in ud1 are identical to the Unit
	 *         objects of ud2, false otherwise.
	 */
	public static boolean areIdentical(UnitDefinition ud1, UnitDefinition ud2) {
		UnitDefinition ud1clone = ud1.clone().simplify();
		UnitDefinition ud2clone = ud2.clone().simplify();
		if (ud1clone.getNumUnits() == ud2clone.getNumUnits()) {
			boolean identical = true;
			for (int i = 0; i < ud1clone.getNumUnits(); i++) {
				identical &= Unit.areIdentical(ud1clone.getUnit(i), ud2clone
						.getUnit(i));
			}
			return identical;
		}
		return false;
	}

	/**
	 * This method returns the predefined unit with the given identifier for the
	 * specified level and version combination or null if either for the given
	 * combination of level and version there is no such predefined unit or the
	 * identifier is not one of those belonging to the group of predefined unit
	 * definitions.
	 * 
	 * @param id
	 *            one of the values
	 *            <ul>
	 *            <li>substance</li>
	 *            <li>volume</li>
	 *            <li>area</li>
	 *            <li>length</li>
	 *            <li>time</li>
	 *            </ul>
	 * @param level
	 *            a number greater than zero.
	 * @param version
	 *            a number greater than zero.
	 * @return The predefined unit definition with the given identifier for the
	 *         specified level version combination or null if no such predefined
	 *         unit exists.
	 */
	public static final UnitDefinition getPredefinedUnit(String id, int level,
			int version) {
		if (level < 3) {
			id = id.toLowerCase();
			Unit u = new Unit(level, version);
			UnitDefinition ud = new UnitDefinition(id, level, version);
			if (id.equals("substance")) {
				u.setKind(Kind.MOLE);
			} else if (id.equals("volume")) {
				u.setKind(Kind.LITRE);
			} else if (id.equals("area")) {
				u.setKind(Kind.METRE);
				u.setExponent(2);
			} else if (id.equals("length")) {
				u.setKind(Kind.METRE);
			} else if (id.equals("time")) {
				u.setKind(Kind.SECOND);
			} else {
				return null;
			}
			if ((level > 1) && (version > 1)) {
				String resource = u.getKind().getUnitOntologyResource();
				if (resource != null) {
					u.setMetaId("meta_" + id);
					u.addCVTerm(new CVTerm(Qualifier.BQB_IS, resource));
				}
			}
			ud.setName("Predefined unit " + id);
			ud.addUnit(u);
			return ud;
		}
		return null;
	}

	/**
	 * Test if the given unit is a predefined unit.
	 * 
	 * @param ud
	 * @deprecated use {@link #isPredefined()}
	 */
	@Deprecated
	public static boolean isBuiltIn(UnitDefinition ud) {
		return isPredefined(ud);
	}

	/**
	 * 
	 * @param name
	 * @param level
	 * @return
	 */
	public static boolean isPredefined(String name, int level) {
		return Unit.isPredefined(name, level);
	}

	/**
	 * Test if the given unit is a predefined unit.
	 * 
	 * @param ud
	 * @return
	 */
	public static boolean isPredefined(UnitDefinition ud) {
		if (ud.getLevel() > 2) {
			return false;
		}
		if (ud.getNumUnits() == 1) {
			UnitDefinition predef = getPredefinedUnit(ud.getId(),
					ud.getLevel(), ud.getVersion());
			if ((predef != null)
					&& Unit.isPredefined(ud.getId(), ud.getLevel())) {
				return ud.equals(predef);
			}
		}
		return false;
	}

	/**
	 * Predefined unit for length.
	 */
	public static final UnitDefinition length(int level, int version) {
		return getPredefinedUnit("length", level, version);
	}

	/**
	 * Returns a string that expresses the unit definition represented by this
	 * UnitDefinition object.
	 * 
	 * @param ud
	 *            the UnitDefinition object
	 * @return a string expressing the unit definition
	 */
	public static String printUnits(UnitDefinition ud) {
		return printUnits(ud, false);
	}

	/**
	 * Returns a string that expresses the unit definition represented by this
	 * UnitDefinition object.
	 * 
	 * @param ud
	 *            the UnitDefinition object
	 * @param compact
	 *            boolean indicating whether the compact form should be used
	 *            (defaults to false)
	 * @return a string expressing the unit definition
	 */
	public static String printUnits(UnitDefinition ud, boolean compact) {
		StringBuilder sb = new StringBuilder();
		if (ud == null) {
			sb.append("null");
		} else {
			for (int i = 0; i < ud.getNumUnits(); i++) {
				Unit unit = ud.getUnit(i);
				if (i > 0) {
					sb.append('*'); // multiplication dot \u22c5.
				}
				if (compact) {
					sb.append(unit.toString());
				} else {
					sb.append(unit.getKind().getName().toLowerCase());
					sb.append(String.format(
							" (exponent = %f, multiplier = %s, scale = %d)",
							unit.getExponent(), StringTools.toString(unit
									.getMultiplier()), unit.getScale()));
				}
			}
		}
		return sb.toString();
	}

	/**
	 * Orders alphabetically the Unit objects within the ListOfUnits of a
	 * UnitDefinition.
	 * 
	 * @param ud
	 *            the UnitDefinition object whose units are to be reordered.
	 */
	public static void reorder(UnitDefinition ud) {
		if (1 < ud.getNumUnits()) {
			ListOf<Unit> orig = ud.getListOfUnits();
			ListOf<Unit> units = new ListOf<Unit>(ud.getLevel(), ud
					.getVersion());
			units.setSBaseListType(Type.listOfUnits);
			orig.removeAllSBaseChangedListeners();
			units.add(orig.remove(orig.size() - 1));
			int i, j;
			for (i = orig.size() - 1; i >= 0; i--) {
				Unit u = orig.remove(i);
				j = 0;
				while (j < units.size()
						&& 0 < u.getKind().compareTo(units.get(j).getKind())) {
					j++;
				}
				units.add(j, u);
			}
			ud.setListOfUnits(units);
		}
	}

	/**
	 * 
	 * @param ud
	 * @return
	 */
	public static UnitDefinition simplify(UnitDefinition ud) {
		return ud.clone().simplify();
	}

	/**
	 * Predefined unit for substance.
	 */
	public static final UnitDefinition substance(int level, int version) {
		return getPredefinedUnit("substance", level, version);
	}

	/**
	 * Predefined unit for time.
	 */
	public static final UnitDefinition time(int level, int version) {
		return getPredefinedUnit("time", level, version);
	}

	/**
	 * Predefined unit for volume.
	 */
	public static final UnitDefinition volume(int level, int version) {
		return getPredefinedUnit("volume", level, version);
	}

	/**
	 * Represents the 'listOfUnit' XML sub-element of a UnitDefinition.
	 */
	private ListOf<Unit> listOfUnits;

	/**
	 * Creates an UnitDefinition instance. By default, the listOfUnit is null.
	 */
	public UnitDefinition() {
		super();
	}

	/**
	 * 
	 * @param level
	 * @param version
	 */
	public UnitDefinition(int level, int version) {
		super(level, version);
	}

	/**
	 * 
	 * @param id
	 */
	public UnitDefinition(String id) {
		super(id);
	}
	
	/**
	 * Creates an UnitDefinition instance from an id, level and version. By
	 * default, the listOfUnit is null.
	 * 
	 * @param id
	 * @param level
	 * @param version
	 */
	public UnitDefinition(String id, int level, int version) {
		super(id, level, version);
	}

	/**
	 * Creates an UnitDefinition instance from an id, level and version. By
	 * default, the listOfUnit is null.
	 * 
	 * @param id
	 * @param name
	 * @param level
	 * @param version
	 */
	public UnitDefinition(String id, String name, int level, int version) {
		super(id, name, level, version);
	}

	/**
	 * Creates an UnitDefinition instance from a given UnitDefinition.
	 * 
	 * @param unitDefinition
	 */
	public UnitDefinition(UnitDefinition unitDefinition) {
		super(unitDefinition);
		if (unitDefinition.isSetListOfUnits()) {
			setListOfUnits(unitDefinition.getListOfUnits().clone());
		}
	}

	/**
	 * 
	 * @param unit
	 */
	public void addUnit(String unit) {
		addUnit(Unit.Kind.valueOf(unit));
	}

	/**
	 * Adds an {@link Unit} to this {@link UnitDefinition}.
	 * 
	 * @param u
	 */
	public void addUnit(Unit u) {
		listOfUnits = getListOfUnits();
		if (!listOfUnits.contains(u)) {
			listOfUnits.add(u);
		}
	}

	/**
	 * Convenient method to add a new unit object with the given kind that will
	 * have the same level/version combination than this {@link UnitDefinition}
	 * object.
	 * 
	 * @param kind
	 */
	public void addUnit(Unit.Kind kind) {
		addUnit(new Unit(kind, getLevel(), getVersion()));
	}

	/**
	 * Removes all {@link Unit} elements from the list of Units in this object.
	 */
	public void clear() {
		getListOfUnits().clear();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sbml.jsbml.element.AbstractSBase#clone()
	 */
	@Override
	public UnitDefinition clone() {
		return new UnitDefinition(this);
	}

	/**
	 * This method converts this unit definition to a
	 */
	public void convertToSIUnits() {
		UnitDefinition ud[] = new UnitDefinition[getNumUnits()];
		Set<SBaseChangedListener> listeners = new HashSet<SBaseChangedListener>(
				getSetOfSBaseChangedListeners());
		removeAllSBaseChangedListeners();
		ListOf<Unit> oldListOfUnits = getListOfUnits().clone();
		for (int i = ud.length - 1; i >= 0; i--) {
			ud[i] = Unit.convertToSI(removeUnit(i));
		}
		for (UnitDefinition u : ud) {
			getListOfUnits().addAll(u.getListOfUnits());
		}
		simplify();
		addAllChangeListeners(listeners);
		firePropertyChange(SBaseChangedEvent.listOfUnits, oldListOfUnits,
				getListOfUnits());
	}

	/**
	 * 
	 * @return
	 */
	public Unit createUnit() {
		return createUnit(Unit.Kind.INVALID);
	}

	/**
	 * 
	 * @param kind
	 * @return
	 */
	public Unit createUnit(Unit.Kind kind) {
		Unit unit = new Unit(kind, getLevel(), getVersion());
		addUnit(unit);

		return unit;
	}

	/**
	 * Devides this unit definition by the second unit definition.
	 * 
	 * @param definition
	 */
	public UnitDefinition divideBy(UnitDefinition definition) {
		if (definition.isSetListOfUnits()) {
			for (Unit unit1 : definition.getListOfUnits()) {
				Unit unit = unit1.clone();
				unit.setExponent(-unit1.getExponent());
				addUnit(unit);
			}
		}
		return this;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sbml.jsbml.element.SBase#equals(Object o)
	 */
	public boolean equals(Object o) {
		if (o instanceof UnitDefinition) {
			UnitDefinition u = (UnitDefinition) o;
			boolean equal = super.equals(o);
			equal &= isSetListOfUnits() == u.isSetListOfUnits();
			if (equal && isSetListOfUnits()) {
				equal &= getListOfUnits().equals(u.getListOfUnits());
			}
			return equal;
		}
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sbml.jsbml.AbstractSBase#getAllowsChildren()
	 */
	@Override
	public boolean getAllowsChildren() {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sbml.jsbml.AbstractSBase#getChildAt(int)
	 */
	@Override
	public ListOf<Unit> getChildAt(int index) {
		int children = getChildCount();
		if (index >= children) {
			throw new IndexOutOfBoundsException(index + " >= " + children);
		}
		int pos = 0;
		if (isSetListOfUnits()) {
			if (index == pos) {
				return getListOfUnits();
			}
			pos++;
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.sbml.jsbml.AbstractSBase#getChildCount()
	 */
	@Override
	public int getChildCount() {
		int children = 0;
		if (isSetListOfUnits()) {
			children++;
		}
		return children;
	}

	/**
	 * 
	 * @return the listOfUnits of this UnitDefinition. Can be empty.
	 */
	public ListOf<Unit> getListOfUnits() {
		if (listOfUnits == null) {
			listOfUnits = ListOf.newInstance(this, Unit.class);
		}
		return listOfUnits;
	}

	/**
	 * 
	 * @return the number of Unit.
	 */
	public int getNumUnits() {
		return isSetListOfUnits() ? listOfUnits.size() : 0;
	}

	/*
	 * (non-Javadoc)
	 * @see org.sbml.jsbml.AbstractSBase#getParent()
	 */
	@SuppressWarnings("unchecked")
	public ListOf<UnitDefinition> getParent() {
		return (ListOf<UnitDefinition>) super.getParent();
	}

	/**
	 * Returns a specific Unit instance belonging to this UnitDefinition.
	 * 
	 * @param i
	 *            an integer, the index of the Unit to be returned.
	 * @return the ith Unit of this UnitDefinition
	 */
	public Unit getUnit(int i) {
		return getListOfUnits().get(i);
	}

	/**
	 * This method tests if this unit definition is a predefined unit.
	 * 
	 * @return
	 * @deprecated use isPredefined()
	 */
	@Deprecated
	public boolean isBuiltIn() {
		return isBuiltIn(this);
	}
	
	/**
	 * This method checks, if this UnitDefinition only contains Invalid
	 * as {@link Unit.Kind}.
	 * 
	 * @return
	 */
	public boolean isInvalid(){
		UnitDefinition ud = this.clone().simplify();
		
		if(ud.getNumUnits() == 1){
			return ud.getUnit(0).isInvalid();
		}
	
		return false; 
		
	}

	/**
	 * This method tests if this unit definition is a predefined unit.
	 * 
	 * @return
	 */
	public boolean isPredefined() {
		return isPredefined(this);
	}

	/**
	 * 
	 * @return true if the listOfUnits is not null and not empty.
	 */
	public boolean isSetListOfUnits() {
		return (listOfUnits != null) && (listOfUnits.size() > 0);
	}

	/**
	 * Convenient method to test whether this {@link UnitDefinition} contains
	 * exactly one {@link Unit} that itself represents a {@link Kind}, i.e.,
	 * multiplier = 1, exponent = 1, scale = 1. Note that this method requires
	 * the level and version attributes of this {@link UnitDefinition} to be
	 * set.
	 * 
	 * @return
	 */
	public boolean isUnitKind() {
		if (getNumUnits() == 1) {
			return Unit.isUnitKind(getUnit(0).getKind(), getLevel(),
					getVersion());
		}
		return false;
	}

	/**
	 * 
	 * @return true if this UnitDefinition is a variant of Area
	 */
	public boolean isVariantOfArea() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 1) {
				Unit unit = listOfUnits.get(0);
				return unit.isVariantOfArea();
			}
		}
		return false;
	}

	/**
	 * Convenience function for testing if a given unit definition is a variant
	 * of the predefined unit identifier 'length'.
	 * 
	 * @param two
	 * @return true if this UnitDefinition is a variant of the predefined unit
	 *         length, meaning metres with only abritrary variations in scale or
	 *         multiplier values; false otherwise.
	 */
	public boolean isVariantOfLength() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 1) {
				Unit unit = listOfUnits.get(0);
				return unit.isVariantOfLength();
			}
		}
		return false;
	}

	/**
	 * Convenience function for testing if a given unit definition is a variant
	 * of the predefined unit identifier 'substance'.
	 * 
	 * @return true if this UnitDefinition is a variant of the predefined unit
	 *         substance, meaning moles or items (and grams or kilograms from
	 *         SBML Level 2 Version 2 onwards) with only abritrary variations in
	 *         scale or multiplier values; false otherwise.
	 */
	public boolean isVariantOfSubstance() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 1) {
				Unit unit = listOfUnits.get(0);
				return unit.isVariantOfSubstance();
			}
		}
		return false;
	}

	/**
	 * 
	 * @return
	 */
	public boolean isVariantOfSubstancePerArea() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 2) {
				if (getUnit(0).isVariantOfSubstance()) {
					Unit two = getUnit(1).clone();
					two.setExponent(-two.getExponent());
					return two.isVariantOfArea();
				} else if (getUnit(1).isVariantOfSubstance()) {
					Unit one = getUnit(0).clone();
					one.setExponent(-one.getExponent());
					return one.isVariantOfArea();
				}
			}
		}
		return false;
	}

	/**
	 * 
	 * @return true if this UnitDefinition is a variant of substance per length.
	 */
	public boolean isVariantOfSubstancePerLength() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 2) {
				Unit unit = listOfUnits.get(0);
				Unit unit2 = listOfUnits.get(1);
				if (unit.isVariantOfSubstance()) {
					Unit two = listOfUnits.get(1).clone();
					two.setExponent(-two.getExponent());
					return two.isVariantOfLength();
				} else if (unit2.isVariantOfSubstance()) {
					Unit one = listOfUnits.get(0).clone();
					one.setExponent(-one.getExponent());
					return one.isVariantOfLength();
				}
			}
		}
		return false;
	}

	/**
	 * 
	 * @return
	 */
	public boolean isVariantOfSubstancePerTime() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 2) {
				Unit unit1 = listOfUnits.get(0);
				Unit unit2 = listOfUnits.get(1);
				if (unit1.isVariantOfSubstance()) {
					Unit two = listOfUnits.get(1).clone();
					two.setExponent(-two.getExponent());
					return two.isVariantOfTime();
				} else if (unit2.isVariantOfSubstance()) {
					Unit one = listOfUnits.get(0).clone();
					one.setExponent(-one.getExponent());
					return one.isVariantOfTime();
				}
			}
		}
		return false;
	}

	/**
	 * 
	 * @return true if this UnitDefinition is a variant of substance per volume.
	 */
	public boolean isVariantOfSubstancePerVolume() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 2) {
				Unit unit = listOfUnits.get(0);
				Unit unit2 = listOfUnits.get(1);
				if (unit.isVariantOfSubstance()) {
					Unit two = listOfUnits.get(1).clone();
					two.setExponent(-two.getExponent());
					return two.isVariantOfVolume();
				} else if (unit2.isVariantOfSubstance()) {
					Unit one = listOfUnits.get(0).clone();
					one.setExponent(-one.getExponent());
					return one.isVariantOfVolume();
				}
			}
		}
		return false;
	}

	/**
	 * Convenience function for testing if a given unit definition is a variant
	 * of the predefined unit identifier 'time'.
	 * 
	 * @return <code>true</code> if this {@link UnitDefinition} is a variant of
	 *         the predefined unit time, meaning second with only arbitrary
	 *         variations in scale or multiplier values; <code>false</code>
	 *         otherwise.
	 */
	public boolean isVariantOfTime() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 1) {
				Unit unit = listOfUnits.get(0);
				return unit.isVariantOfTime();
			}
		}
		return false;
	}
	
	/**
	 * Convenience function for testing if a given unit definition is a variant
	 * of the predefined unit identifier 'volume'.
	 * 
	 * @return <code>true</code> if this {@link UnitDefinition} is a variant of
	 *         the predefined unit volume, meaning litre or cubic metre with
	 *         only arbitrary variations in scale or multiplier values;
	 *         <code>false</code> otherwise.
	 */
	public boolean isVariantOfVolume() {
		if (isSetListOfUnits()) {
			if (listOfUnits.size() == 1) {
				Unit unit = listOfUnits.get(0);
				return unit.isVariantOfVolume();
			}
		}
		return false;
	}

	/**
	 * Multiplies this unit with the given unit definition, i.e., adds a clone
	 * of each unit object in the list of units of the given definition to the
	 * list of this unit.
	 * 
	 * @param definition
	 * @return
	 */
	public UnitDefinition multiplyWith(UnitDefinition definition) {
		if (!isSetListOfUnits()) {
			this.listOfUnits = new ListOf<Unit>(getLevel(), getVersion());
		}
		for (Unit unit : definition.getListOfUnits()) {
			addUnit(unit.clone());
		}
		return this;
	}

	/**
	 * Raises this unit definition by the power of the given exponent, i.e., the
	 * exponents of every unit contained by this unit definition are multiplied
	 * with the given exponent.
	 * 
	 * @param exponent
	 * @return a pointer to this UnitDefinition.
	 */
	public UnitDefinition raiseByThePowerOf(int exponent) {
		if (isSetListOfUnits()) {
			Unit u;
			for (int i = listOfUnits.size() - 1; i >= 0; i--) {
				u = listOfUnits.get(i);
				u.setExponent(u.getExponent() * exponent);
				if (u.getExponent() == 0) {
					listOfUnits.remove(i);
				}
			}
		}
		return this;
	}
	
	/**
	 * Removes the nth Unit object from this UnitDefinition object and returns a
	 * pointer to it.
	 * 
	 * The caller owns the returned object and is responsible for deleting it.
	 * 
	 * @param i
	 *            the index of the Unit object to remove
	 * @return the Unit object removed. As mentioned above, the caller owns the
	 *         returned item. NULL is returned if the given index is out of
	 *         range.
	 */
	public Unit removeUnit(int i) {
		if (isSetListOfUnits()) {
			Unit u = listOfUnits.remove(i);
			if (u != null) {
				u.fireSBaseRemovedEvent();
			}
			return u;
		}
		return null;
	}

	/**
	 * Sets the {@link #listOfUnits} of this {@link UnitDefinition}.
	 * Automatically sets the parent SBML object of the list to this
	 * {@link UnitDefinition} instance.
	 * 
	 * @param listOfUnits
	 */
	public void setListOfUnits(ListOf<Unit> listOfUnits) {
		unsetListOfUnits();
		this.listOfUnits = listOfUnits;
		if ((this.listOfUnits != null) && (this.listOfUnits.getSBaseListType() != ListOf.Type.listOfUnits)) {
			this.listOfUnits.setSBaseListType(ListOf.Type.listOfUnits);
		}
		setThisAsParentSBMLObject(this.listOfUnits);
	}
	
	/**
	 * Simplifies the {@link UnitDefinition} so that any {@link Unit} objects
	 * occurring within the {@link #listOfUnits} occurs only once. {@link Unit}s
	 * of {@link Kind} {@link Kind.INVALID} are treated like
	 * {@link Kind.DIMENSIONLESS} units and will therefore tend to disapear by
	 * merging with other units.
	 * 
	 * @return a pointer to the simplified {@link UnitDefinition}.
	 */
	public UnitDefinition simplify() {
		if (isSetListOfUnits()) {
			reorder(this);
			for (int i = getNumUnits() - 2; i >= 0; i--) {
				Unit u = getUnit(i); // current unit
				Unit s = getUnit(i + 1); // successor unit
				if (Unit.Kind.areEquivalent(u.getKind(), s.getKind())
						|| u.isDimensionless() || s.isDimensionless() || u.isInvalid() || s.isInvalid()) {
					Unit.merge(u, removeUnit(i + 1));
					if (u.isDimensionless() && (i == 0) && (getNumUnits() > 1)) {
						Unit.merge(getUnit(i + 1), removeUnit(i));
					}
				}
			}
		}
		return this;
	}

	/**
	 * Removes the {@link #listOfUnits} from this {@link UnitDefinition} and notifies 
	 * all registered instances of {@link SBaseChangedListener}.
	 */
	public void unsetListOfUnits() {
		if (this.listOfUnits != null) {
			ListOf<Unit> oldListOfUnits = this.listOfUnits;
			this.listOfUnits = null;
			oldListOfUnits.fireSBaseRemovedEvent();
		}
	}

}
